from datetime import datetime

from dateutil.rrule import MONTHLY, YEARLY, rrule

from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.scraping.filter import inverse, regex_filter
from fundus.scraping.url import NewsMap, RSSFeed, Sitemap

from ..shared import EuronewsParser
from .berliner_zeitung import BerlinerZeitungParser
from .bild import BildParser
from .boersenzeitung import BoersenZeitungParser
from .br import BRParser
from .business_insider_de import BusinessInsiderDEParser
from .die_welt import DieWeltParser
from .die_zeit import DieZeitParser
from .dw import DWParser
from .faz import FAZParser
from .focus import FocusParser
from .frankfurter_rundschau import FrankfurterRundschauParser
from .freiepresse import FreiePresseParser
from .funke import FunkeParser
from .gamestar import GamestarParser
from .golem import GolemParser
from .heise import HeiseParser
from .hessenschau import HessenschauParser
from .junge_welt import JungeWeltParser
from .kicker import KickerParser
from .krautreporter import KrautreporterParser
from .mdr import MDRParser
from .merkur import MerkurParser
from .motorsport_magazin import MotorSportMagazinParser
from .mz import MitteldeutscheZeitungParser
from .ndr import NDRParser
from .netzpolitik_org import NetzpolitikOrgParser
from .ntv import NTVParser
from .postillon import PostillonParser
from .rheinische_post import RheinischePostParser
from .rn import RuhrNachrichtenParser
from .spon import SPONParser
from .sportschau import SportSchauParser
from .stern import SternParser
from .sz import SZParser
from .tagesschau import TagesschauParser
from .tagesspiegel import TagesspiegelParser
from .taz import TazParser
from .vogue_de import VogueDEParser
from .waz import WAZParser
from .wdr import WDRParser
from .winfuture import WinfutureParser
from .zdf import ZDFParser


# noinspection PyPep8Naming
class DE(metaclass=PublisherGroup):
    default_language = "de"

    SportSchau = Publisher(
        name="Sportschau",
        domain="https://www.sportschau.de/",
        parser=SportSchauParser,
        sources=[
            RSSFeed("https://www.sportschau.de/index~rss2.xml"),
            Sitemap("https://www.sportschau.de/index~sitemap_p-0.xml"),
            NewsMap("https://www.sportschau.de/kompakt-sp-100~news.xml"),
        ],
        url_filter=inverse(regex_filter("sportschau.de")),
    )

    NetzpolitikOrg = Publisher(
        name="netzpolitik.org",
        domain="https://netzpolitik.org/",
        parser=NetzpolitikOrgParser,
        sources=[
            Sitemap(
                "https://netzpolitik.org/sitemap.xml",
                sitemap_filter=inverse(regex_filter("sitemap-posttype-post")),
            ),
            RSSFeed("https://netzpolitik.org/feed/"),
        ],
    )

    BerlinerMorgenpost = Publisher(
        name="Berliner Morgenpost",
        domain="https://www.morgenpost.de/",
        parser=FunkeParser,
        sources=[NewsMap("https://www.morgenpost.de/sitemaps/news.xml")]
        + [
            Sitemap(
                f"https://www.morgenpost.de/sitemaps/archive/sitemap-{d.year}-{str(d.month).zfill(2)}-p00.xml.gz",
            )
            for d in reversed(list(rrule(MONTHLY, dtstart=datetime(2003, 2, 1), until=datetime.now())))
        ],
    )

    HamburgerAbendblatt = Publisher(
        name="Hamburger Abendblatt",
        domain="https://www.abendblatt.de/",
        parser=FunkeParser,
        sources=[
            RSSFeed("https://www.abendblatt.de/rss"),
            NewsMap("https://www.abendblatt.de/sitemaps/news.xml"),
        ]
        + [
            Sitemap(
                f"https://www.abendblatt.de/sitemaps/archive/sitemap-{d.year}-{str(d.month).zfill(2)}-p00.xml.gz",
            )
            for d in reversed(list(rrule(MONTHLY, dtstart=datetime(2000, 4, 1), until=datetime.today())))
        ],
    )

    DieWelt = Publisher(
        name="Die Welt",
        domain="https://www.welt.de/",
        parser=DieWeltParser,
        sources=[
            RSSFeed("https://www.welt.de/feeds/latest.rss"),
            Sitemap("https://www.welt.de/sitemaps/sitemap/sitemap.xml"),
            NewsMap("https://www.welt.de/sitemaps/newssitemap/newssitemap.xml"),
        ],
        url_filter=regex_filter("/Anlegertipps-|/videos?[0-9]{2}|/mediathek/"),
    )

    MDR = Publisher(
        name="Mitteldeutscher Rundfunk (MDR)",
        domain="https://www.mdr.de/",
        parser=MDRParser,
        sources=[
            RSSFeed("https://www.mdr.de/nachrichten/index-rss.xml"),
            Sitemap("https://www.mdr.de/sitemap-index-100.xml"),
            NewsMap("https://www.mdr.de/news-sitemap.xml"),
        ],
    )

    FAZ = Publisher(
        name="Frankfurter Allgemeine Zeitung",
        domain="https://www.faz.net/",
        parser=FAZParser,
        sources=[
            RSSFeed("https://www.faz.net/rss/aktuell"),
            RSSFeed("https://www.faz.net/rss/aktuell/politik"),
            RSSFeed("https://www.faz.net/rss/aktuell/sport"),
            RSSFeed("https://www.faz.net/rss/aktuell/wirtschaft/"),
            RSSFeed("https://www.faz.net/rss/aktuell/gesellschaft/"),
            Sitemap("https://www.faz.net/sitemap-index.xml", sitemap_filter=inverse(regex_filter("-artikel-"))),
            NewsMap("https://www.faz.net/sitemap-news.xml"),
        ],
    )

    Focus = Publisher(
        name="Focus Online",
        domain="https://www.focus.de/",
        parser=FocusParser,
        sources=[
            NewsMap("https://www.focus.de/sitemap_news_ressorts.xml"),
        ],
        request_header={"user-agent": "Fundus"},
    )

    Merkur = Publisher(
        name="Münchner Merkur",
        domain="https://www.merkur.de/",
        parser=MerkurParser,
        sources=[
            RSSFeed("https://www.merkur.de/welt/rssfeed.rdf"),
            Sitemap("https://www.merkur.de/sitemap-index.xml"),
            NewsMap("https://www.merkur.de/news.xml"),
        ],
    )

    SZ = Publisher(
        name="Süddeutsche Zeitung",
        domain="https://www.sueddeutsche.de/",
        parser=SZParser,
        sources=[RSSFeed("https://rss.sueddeutsche.de/alles")],
    )

    SpiegelOnline = Publisher(
        name="Spiegel Online",
        domain="https://www.spiegel.de/",
        parser=SPONParser,
        sources=[
            RSSFeed("https://www.spiegel.de/schlagzeilen/index.rss"),
            Sitemap("https://www.spiegel.de/sitemap.xml"),
            NewsMap("https://www.spiegel.de/sitemaps/news-de.xml"),
        ],
        request_header={"User-Agent": "Googlebot"},
    )

    DieZeit = Publisher(
        name="Die Zeit",
        domain="https://www.zeit.de/",
        parser=DieZeitParser,
        sources=[
            RSSFeed("https://newsfeed.zeit.de/news/index"),
            Sitemap("https://www.zeit.de/gsitemaps/index.xml", reverse=True),
            NewsMap(
                f"https://www.zeit.de/gsitemaps/index.xml?date={datetime.now().strftime('%Y-%m-%d')}&unit=days&period=1",
            ),
        ],
        url_filter=regex_filter(
            "/zett/|/angebote/|/kaenguru-comics/|/administratives/|/index(?!.)|/elbvertiefung-[0-9]{2}-[0-9]{2}"
        ),
        request_header={"user-agent": "Googlebot"},
    )

    BerlinerZeitung = Publisher(
        name="Berliner Zeitung",
        domain="https://www.berliner-zeitung.de/",
        parser=BerlinerZeitungParser,
        sources=[
            RSSFeed("https://www.berliner-zeitung.de/feed.xml"),
            Sitemap("https://www.berliner-zeitung.de/sitemap.xml"),
            NewsMap("https://www.berliner-zeitung.de/news-sitemap.xml"),
        ],
        url_filter=regex_filter("/news/"),
    )

    Tagesschau = Publisher(
        name="Tagesschau",
        domain="https://www.tagesschau.de/",
        parser=TagesschauParser,
        sources=[RSSFeed("https://www.tagesschau.de/xml/rss2/")],
    )

    DW = Publisher(
        name="Deutsche Welle",
        domain="https://www.dw.com/",
        parser=DWParser,
        sources=[
            RSSFeed("https://rss.dw.com/xml/rss-de-all"),
            Sitemap("https://www.dw.com/de/article-sitemap.xml"),
            NewsMap("https://www.dw.com/de/news-sitemap.xml"),
            RSSFeed("https://rss.dw.com/xml/rss-en-all", languages={"en"}),
            Sitemap("https://www.dw.com/en/article-sitemap.xml", languages={"en"}),
            NewsMap("https://www.dw.com/en/news-sitemap.xml", languages={"en"}),
        ],
    )
    Stern = Publisher(
        name="Stern",
        domain="https://www.stern.de/",
        parser=SternParser,
        sources=[RSSFeed("https://www.stern.de/feed/standard/alle-nachrichten/")],
    )

    NTV = Publisher(
        name="N-Tv",
        domain="https://www.n-tv.de/",
        parser=NTVParser,
        sources=[
            NewsMap("https://www.n-tv.de/news.xml"),
            Sitemap("https://www.n-tv.de/sitemap.xml", sitemap_filter=regex_filter("sitemap-sections")),
        ],
    )

    NDR = Publisher(
        name="Norddeutscher Rundfunk (NDR)",
        domain="https://www.ndr.de/",
        parser=NDRParser,
        sources=[
            NewsMap("https://www.ndr.de/news-102~news.xml"),
            Sitemap("https://www.ndr.de/index~sitemap_f-sitemap--ndr--info--100.xml"),
            Sitemap("https://www.ndr.de/index~sitemap_f-sitemap--ndr--niedersachsen--100.xml"),
            Sitemap("https://www.ndr.de/index~sitemap_f-sitemap--ndr--schleswig--holstein--100.xml"),
            Sitemap("https://www.ndr.de/index~sitemap_f-sitemap--ndr--mecklenburg--vorpommern--100.xml"),
            Sitemap("https://www.ndr.de/index~sitemap_f-sitemap--ndr--hamburg--100.xml"),
            Sitemap("https://www.ndr.de/index~sitemap_f-sitemap--ndr--sport--100.xml"),
            Sitemap("https://www.ndr.de/index~sitemap_f-sitemap--ndr--kultur--100.xml"),
            Sitemap("https://www.ndr.de/index~sitemap_f-sitemap--ndr--geschichte--100.xml"),
        ],
        url_filter=regex_filter("podcast[0-9]{4}|/index.html"),
    )

    Taz = Publisher(
        name="Die Tageszeitung (taz)",
        domain="https://taz.de/",
        parser=TazParser,
        sources=[
            NewsMap("https://taz.de/sitemap-google-news.xml"),
            Sitemap("https://taz.de/sitemap-index.xml"),
        ],
    )

    Heise = Publisher(
        name="Heise",
        domain="https://www.heise.de",
        sources=[
            RSSFeed("https://www.heise.de/rss/heise.rdf"),
            Sitemap(
                "https://www.heise.de/sitemapindex.xml",
                sitemap_filter=inverse(regex_filter("/news/")),
            ),
        ],
        parser=HeiseParser,
        query_parameter={"seite": "all"},
    )

    Bild = Publisher(
        name="Bild",
        domain="https://www.bild.de/",
        parser=BildParser,
        sources=[
            RSSFeed("https://www.bild.de/feed/alles.xml"),
            NewsMap("https://www.bild.de/sitemap-news.xml"),
            Sitemap("https://www.bild.de/sitemap-index.xml"),
        ],
    )

    WAZ = Publisher(
        name="Westdeutsche Allgemeine Zeitung (WAZ)",
        domain="https://www.waz.de/",
        parser=WAZParser,
        sources=[NewsMap("https://www.waz.de/sitemaps/news.xml")],
    )

    BSZ = Publisher(
        name="Braunschweiger Zeitung",
        domain="https://www.braunschweiger-zeitung.de/",
        parser=FunkeParser,
        sources=[
            RSSFeed("https://www.braunschweiger-zeitung.de/rss"),
            NewsMap("https://www.braunschweiger-zeitung.de/sitemaps/news.xml"),
        ]
        + [
            Sitemap(
                f"https://www.braunschweiger-zeitung.de/sitemaps/archive/sitemap-{d.year}-{str(d.month).zfill(2)}-p00.xml.gz",
            )
            for d in list(rrule(MONTHLY, dtstart=datetime(2005, 12, 1), until=datetime.now()))
        ],
    )

    BusinessInsiderDE = Publisher(
        name="Business Insider DE",
        domain="https://www.businessinsider.de/",
        parser=BusinessInsiderDEParser,
        sources=[
            RSSFeed("https://www.businessinsider.de/feed/businessinsider-alle-artikel"),
            NewsMap("https://www.businessinsider.de/news-sitemap.xml"),
            Sitemap(
                "https://www.businessinsider.de/sitemap_index.xml",
                sitemap_filter=inverse(regex_filter("post-sitemap")),
                reverse=True,
            ),
        ],
    )

    RheinischePost = Publisher(
        name="Rheinische Post",
        domain="https://rp-online.de/",
        parser=RheinischePostParser,
        sources=[
            RSSFeed("https://rp-online.de/feed.rss"),
            NewsMap("https://rp-online.de/sitemap-news.xml"),
            Sitemap("https://rp-online.de/sitemap.xml"),
        ],
    )

    Golem = Publisher(
        name="Golem",
        domain="https://www.golem.de/",
        sources=[
            RSSFeed("https://www.golem.de/rss"),
            NewsMap("https://www.golem.de/news/gsitemap-2404.xml"),
            Sitemap("https://www.golem.de/gsiteindex.xml"),
        ],
        request_header={"User-Agent": "Googlebot"},
        parser=GolemParser,
    )

    WinFuture = Publisher(
        name="WinFuture",
        domain="https://winfuture.de/",
        parser=WinfutureParser,
        sources=[
            RSSFeed("https://static.winfuture.de/feeds/WinFuture-News-rss2.0.xml"),
            NewsMap("https://winfuture.de/sitemap-latest-news.xml.gz"),
            Sitemap(
                "https://winfuture.de/sitemap.xml",
                sitemap_filter=inverse(regex_filter("sitemap-news")),
            ),
        ],
        url_filter=regex_filter("https:////winfuture/.de//news*"),
    )

    JungeWelt = Publisher(
        name="Junge Welt",
        domain="https://www.jungewelt.de/",
        parser=JungeWeltParser,
        sources=[
            RSSFeed("https://www.jungewelt.de/feeds/newsticker.rss"),
        ],
    )

    Tagesspiegel = Publisher(
        name="Tagesspiegel",
        domain="https://www.tagesspiegel.de/",
        parser=TagesspiegelParser,
        sources=[
            NewsMap("https://www.tagesspiegel.de/news.xml"),
        ]
        + [
            Sitemap(f"https://www.tagesspiegel.de/contentexport/static/sitemap-index_{date.year}.xml")
            for date in reversed(list(rrule(YEARLY, dtstart=datetime(1996, 1, 1), until=datetime.today())))
        ],
    )

    EuronewsDE = Publisher(
        name="Euronews (DE)",
        domain="https://de.euronews.com/",
        parser=EuronewsParser,
        sources=[
            Sitemap("https://de.euronews.com/sitemaps/de/articles.xml"),
            NewsMap("https://de.euronews.com/sitemaps/de/latest-news.xml"),
        ],
        url_filter=regex_filter("/video/"),
    )

    Hessenschau = Publisher(
        name="Hessenschau",
        domain="https://www.hessenschau.de/",
        parser=HessenschauParser,
        sources=[
            RSSFeed("https://www.hessenschau.de/index.rss"),
            Sitemap("https://www.hessenschau.de/indexsitemap.nc.xml"),
            Sitemap("https://www.hessenschau.de/sitemap.nc.xml"),
        ],
    )

    WDR = Publisher(
        name="Westdeutscher Rundfunk",
        domain="https://www1.wdr.de/",
        parser=WDRParser,
        sources=[RSSFeed("https://www1.wdr.de/uebersicht-100.feed")],
        url_filter=inverse(regex_filter("wdr.de/(?!mediathek/)")),
    )

    BR = Publisher(
        name="Bayerischer Rundfunk (BR)",
        domain="https://www.br.de/",
        parser=BRParser,
        sources=[
            Sitemap("https://www.br.de/sitemapIndex.xml"),
            NewsMap("https://www.br.de/nachrichten/sitemaps/news.xml"),
        ],
    )

    ZDF = Publisher(
        name="ZDF",
        domain="https://www.zdf.de/",
        parser=ZDFParser,
        sources=[
            Sitemap("https://www.zdf.de/sitemap.xml", reverse=True),
            NewsMap("https://www.zdf.de/news-sitemap.xml"),
            RSSFeed("https://www.zdf.de/rss/zdf/nachrichten"),
        ],
    )

    MotorSportMagazin = Publisher(
        name="MotorSport Magazin",
        domain="https://www.motorsport-magazin.com/",
        parser=MotorSportMagazinParser,
        sources=[
            RSSFeed("https://www.motorsport-magazin.com/rss/alle-rennserien.xml"),
            Sitemap("https://www.motorsport-magazin.com/sitemap.xml"),
        ],
    )

    Postillon = Publisher(
        name="Postillon",
        domain="https://www.der-postillon.com/",
        parser=PostillonParser,
        sources=[
            RSSFeed("https://follow.it/der-postillon-abo"),
            Sitemap("https://www.der-postillon.com/sitemap.xml"),
        ],
        url_filter=regex_filter("https://follow.it/"),
    )

    Kicker = Publisher(
        name="Kicker",
        domain="https://www.kicker.de/",
        parser=KickerParser,
        sources=[
            RSSFeed("https://newsfeed.kicker.de/news/aktuell"),
            Sitemap(
                "https://leserservice.kicker.de/sitemap_0.xml",
                sitemap_filter=regex_filter("leserservice.kicker.de"),
            ),
            NewsMap("https://newsfeed.kicker.de/googlesitemapnews.xml"),
        ],
        url_filter=regex_filter("/slideshow|/video|heute-live|live-konferenz|/bilder|/ticker"),
    )

    Krautreporter = Publisher(
        name="Krautreporter",
        domain="https://krautreporter.de/",
        parser=KrautreporterParser,
        sources=[
            # NOTE: robots.txt mentions that it reserves the right of use for text & data mining (§ 44 b UrhG),
            # but this is not in machine readable format, which is required by law for it to be effective.
            # NOTE: Unfortunately, both sitemap.xml and news.xml are identical.
            Sitemap("https://krautreporter.de/sitemap.xml", reverse=True),
            # NewsMap("https://krautreporter.de/news.xml"),
            RSSFeed("https://krautreporter.de/feeds.rss"),
        ],
        url_filter=regex_filter(r"/(pages|archiv|serien|thema|zusammenhaenge)/"),
    )

    FrankfurterRundschau = Publisher(
        name="Frankfurter Rundschau",
        domain="https://www.fr.de",
        parser=FrankfurterRundschauParser,
        sources=[
            RSSFeed("https://fr.de/rssfeed.rdf"),
            Sitemap("https://www.fr.de/sitemap-index.xml"),
            NewsMap("https://www.fr.de/news.xml"),
        ],
    )

    BoersenZeitung = Publisher(
        name="Börsen-Zeitung",
        domain="https://www.boersen-zeitung.de",
        parser=BoersenZeitungParser,
        sources=[
            NewsMap("https://www.boersen-zeitung.de/sitemap/news.xml.gz"),
            Sitemap(
                "https://www.boersen-zeitung.de/sitemap/index.xml.gz",
                sitemap_filter=regex_filter("/sitemap-0.xml.gz"),
            ),
        ],
    )

    VogueDE = Publisher(
        name="Vogue",
        domain="https://www.vogue.de/",
        parser=VogueDEParser,
        sources=[
            RSSFeed("https://www.vogue.de/feed/rss"),
            NewsMap("https://www.vogue.de/feed/sitemap-news/sitemap-google-news"),
            Sitemap("https://www.vogue.de/sitemap.xml"),
        ],
    )

    MitteldeutscheZeitung = Publisher(
        name="Mitteldeutsche Zeitung",
        domain="https://www.mz.de/",
        parser=MitteldeutscheZeitungParser,
        sources=[
            Sitemap("https://www.mz.de/sitemaps/sitemap-ressort-index.xml"),
            NewsMap("https://www.mz.de/sitemaps/newssitemap-index.xml"),
        ],
    )

    FreiePresse = Publisher(
        name="FreiePresse",
        domain="https://www.freiepresse.de/",
        parser=FreiePresseParser,
        sources=[
            RSSFeed("https://www.freiepresse.de/rss/rss_chemnitz.php"),
            RSSFeed("https://www.freiepresse.de/rss/rss_erzgebirge.php"),
            RSSFeed("https://www.freiepresse.de/rss/rss_mittelsachsen.php"),
            RSSFeed("https://www.freiepresse.de/rss/rss_vogtland.php"),
            RSSFeed("https://www.freiepresse.de/rss/rss_zwickau.php"),
            RSSFeed("https://www.freiepresse.de/rss/rss_politik.php"),
            RSSFeed("https://www.freiepresse.de/rss/rss_wirtschaft.php"),
            RSSFeed("https://www.freiepresse.de/rss/rss_kultur.php"),
            RSSFeed("https://www.freiepresse.de/rss/rss_sport.php"),
            RSSFeed("https://www.freiepresse.de/rss/rss_sachsen.php"),
            RSSFeed("https://www.freiepresse.de/rss/rss_regional.php"),
            Sitemap("https://www.freiepresse.de/sitemaps/articles_last2years.xml", reverse=True),
        ],
    )

    RuhrNachrichten = Publisher(
        name="Ruhr Nachrichten",
        domain="https://www.ruhrnachrichten.de/",
        parser=RuhrNachrichtenParser,
        sources=[
            RSSFeed("https://www.ruhrnachrichten.de/service/feed/"),
            NewsMap("https://www.ruhrnachrichten.de/news-sitemap.xml"),
            Sitemap(
                "https://www.ruhrnachrichten.de/sitemap_index.xml",
                reverse=True,
                sitemap_filter=inverse(regex_filter("post-sitemap")),
            ),
        ],
    )

    Gamestar = Publisher(
        name="Gamestar",
        domain="https://www.gamestar.de/",
        parser=GamestarParser,
        sources=[
            NewsMap("https://www.gamestar.de/sitemapnews.xml"),
            Sitemap("https://www.gamestar.de/artikel_archiv_index.xml"),
        ],
    )



import datetime
import re
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    image_extraction,
)


class BoersenZeitungParser(ParserProxy):
    class V1(BaseParser):
        VALID_UNTIL = datetime.date(2024, 12, 9)

        _paragraph_selector = CSSSelector("storefront-content-body .no-tts p")
        _subheadline_selector = XPath("//p[contains(@class, 'interline')]")
        _summary_selector = CSSSelector("storefront-html.excerpt > div")

        _topic_selector = CSSSelector("a[href^='/thema'] > span")
        _paywall_selector = CSSSelector("storefront-html.paywall-headline > div")

        _title_bloat_pattern = re.compile(r"\|.*")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def title(self) -> Optional[str]:
            if fulltitle := self.precomputed.meta.get("og:title"):
                return re.sub(self._title_bloat_pattern, "", fulltitle).strip()
            return None

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.meta.get("twitter:misc:Written by"))

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.meta.get("article:published:time"))

        @attribute
        def topics(self) -> List[str]:
            return [node.text_content().strip() for node in self._topic_selector(self.precomputed.doc)]

        @attribute
        def free_access(self) -> bool:
            # print(self._paywall_selector(self.precomputed.doc).text_content().strip())
            return not [node.text_content().strip() for node in self._paywall_selector(self.precomputed.doc)]

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=XPath("//h1|//script"),
                image_selector=XPath("//storefront-image|//figure//img"),
                author_selector=XPath("./ancestor::storefront-section//storefront-html[@class='image-copyright']"),
            )

    class V1_1(V1):
        VALID_UNTIL = datetime.date.today()

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.meta.get("twitter:data1"))

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.meta.get("article:published_time"))



import re
from datetime import datetime
from typing import List, Optional, Pattern

import lxml.html
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, ParserProxy, attribute
from fundus.parser.data import Image
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_nodes_to_text,
    generic_topic_parsing,
    image_extraction,
)


class KommersantParser(ParserProxy):
    class V1(BaseParser):
        _summary_selector = XPath("//header/h2")
        _paragraph_selector = XPath(
            "//div[contains(@class, 'article_text_wrapper')]/p[not(contains(@class, 'document_authors') or (not(text()) and b))]"
        )

        _author_selector = XPath("//p[@class='doc__text document_authors']")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                summary_selector=self._summary_selector,
            )

        @attribute
        def publishing_date(self) -> Optional[datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.bf_search("headline")

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(generic_nodes_to_text(self._author_selector(self.precomputed.doc)))

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.meta.get("keywords"))

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=XPath("//header/h1"),
                image_selector=XPath("//figure//img[not(contains(@class, 'fallback'))]"),
                caption_selector=XPath("./ancestor::figure//figcaption/p"),
                author_selector=re.compile(r"(?i)Фото:(?P<credits>.+)"),
            )



from collections import defaultdict
from textwrap import indent
from typing import Dict, Iterator, List, Optional, Set, Type, Union
from urllib.robotparser import RobotFileParser
from warnings import warn

import more_itertools
from requests.exceptions import ConnectionError, HTTPError, ReadTimeout
from typing_extensions import TypeAlias

from fundus.logging import create_logger
from fundus.parser.base_parser import ParserProxy
from fundus.scraping.filter import URLFilter
from fundus.scraping.session import session_handler
from fundus.scraping.url import NewsMap, RSSFeed, Sitemap, URLSource
from fundus.utils.iteration import iterate_all_subclasses

logger = create_logger(__name__)

FilterResult: TypeAlias = Union[List["FilteredPublisher"], List["Publisher"]]


class CustomRobotFileParser(RobotFileParser):
    """Monkey patch RobotFileParse

    This class overwrites the read() methode of RobotFileParser to use the <requests> pkg instead of urllib.
    This is in order to avoid 403 errors when fetching the robots.txt file.
    """

    # noinspection PyAttributeOutsideInit
    def read(self, headers: Optional[Dict[str, str]] = None) -> None:
        """Reads the robots.txt URL and feeds it to the parser."""
        try:
            # noinspection PyUnresolvedReferences
            session = session_handler.get_session()
            response = session.get_with_interrupt(self.url, headers=headers)  # type: ignore[attr-defined]
        except HTTPError as err:
            if err.response.status_code in (401, 403):
                self.disallow_all = True
            elif 400 <= err.response.status_code < 500:
                self.allow_all = True
        else:
            self.parse(response.text.splitlines())


class Robots:
    def __init__(self, url: str):
        self.url = url
        self.robots_file_parser = CustomRobotFileParser(url)
        self.ready: bool = False

    def read(self, headers: Optional[Dict[str, str]] = None) -> None:
        try:
            self.robots_file_parser.read(headers=headers)
        except (ConnectionError, ReadTimeout):
            logger.warning(f"Could not load robots {self.url!r}. Ignoring robots and continuing.")
            self.robots_file_parser.allow_all = True
        self.ready = True

    def can_fetch(self, useragent: str, url: str) -> bool:
        return self.robots_file_parser.can_fetch(useragent, url)

    def crawl_delay(self, useragent: str) -> Optional[float]:
        delay = self.robots_file_parser.crawl_delay(useragent)
        return delay if delay is None else float(delay)


class Publisher:
    __name__: str
    __group__: "PublisherGroup"

    __SOURCE_ORDER__: Dict[Type[URLSource], int] = {
        RSSFeed: 1,
        NewsMap: 2,
        Sitemap: 3,
    }

    def __init__(
        self,
        name: str,
        domain: str,
        parser: Type[ParserProxy],
        sources: List[URLSource],
        query_parameter: Optional[Dict[str, str]] = None,
        url_filter: Optional[URLFilter] = None,
        request_header: Optional[Dict[str, str]] = None,
        deprecated: bool = False,
        suppress_robots: bool = False,
    ):
        """Initialization of a new Publisher object

        Args:
            name (str): Name of the publisher, as it would appear on the website
            domain (str): The domain of the publishers website
            parser (Type[ParserProxy]): Corresponding ParserProxy Object
            sources (List[URLSource]): List of sources for articles from the publishers
            query_parameter (Optional[Dict[str, str]]): Dictionary of query parameter: content to be
                appended to crawled URLs
            url_filter (Optional[URLFilter]): Regex filter to apply determining URLs to be skipped
            request_header (Optional[Dict[str, str]]): Request header to be used for the GET-request

        """
        if not (name and domain and parser and sources):
            raise ValueError("Failed to create Publisher. Name, Domain, Parser and Sources are mandatory")
        self.name = name
        self.parser = parser()
        self.domain = domain
        self.query_parameter = query_parameter
        self.url_filter = url_filter
        self.request_header = request_header
        self.deprecated = deprecated
        self.robots = Robots(self.domain + "robots.txt" if self.domain.endswith("/") else self.domain + "/robots.txt")

        # Temporary fix to compensate for a bug in the RobotsFileParser treating rule lines
        # like /? as / disallowing the entire site. we could think about replacing the urllib
        # implementation with https://github.com/seomoz/reppy
        if suppress_robots:
            self.robots.robots_file_parser.allow_all = True

        source_mapping: Dict[Type[URLSource], List[URLSource]] = defaultdict(list)

        for url_source in sources:
            if not isinstance(url_source, URLSource):
                raise TypeError(
                    f"Unexpected type {type(url_source).__name__!r} as source for {self!r}. "
                    f"Allowed are {', '.join(repr(cls.__name__) for cls in iterate_all_subclasses(URLSource))}"
                )
            source_mapping[type(url_source)].append(url_source)

        self._source_mapping = dict(sorted(source_mapping.items(), key=lambda item: self.__SOURCE_ORDER__[item[0]]))

    @property
    def source_mapping(self) -> Dict[Type[URLSource], List[URLSource]]:
        return self._source_mapping

    @property
    def languages(self) -> Set[str]:
        return set.union(*(source.languages for sources in self.source_mapping.values() for source in sources))

    @property
    def source_types(self) -> Set[Type[URLSource]]:
        return set(self.source_mapping.keys())

    def __str__(self) -> str:
        return f"{self.name}"

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Publisher):
            return False
        return (
            self.name == other.name
            and self.parser == other.parser
            and self.domain == other.domain
            and self.source_mapping == other.source_mapping
            and self.query_parameter == other.query_parameter
            and self.url_filter == other.url_filter
            and self.request_header == other.request_header
        )

    def supports(
        self, source_types: Optional[List[Type[URLSource]]] = None, languages: Optional[List[str]] = None
    ) -> bool:
        # we iterate source by source instead of checking self.languages and self.source_types,
        # because we need to know if there is a source supporting the given combination of
        # <source_types> and <languages>
        filtered_sources = [
            source
            for source in more_itertools.flatten(self.source_mapping.values())
            if (type(source) in source_types if source_types else True)
            and (source.languages & set(languages) if languages else True)
        ]

        return any(filtered_sources)


class FilteredPublisher(Publisher):
    """Publisher with prefiltered sources.

    Publisher with attached source types and languages to pre-filter sources.
    """

    def __init__(
        self, source_types: Optional[Set[Type[URLSource]]] = None, languages: Optional[Set[str]] = None, *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.__source_types_filter__ = source_types or set()
        self.__language_filter__ = languages or set()

    @classmethod
    def from_publisher(
        cls,
        publisher: Publisher,
        source_types: Optional[Set[Type[URLSource]]] = None,
        languages: Optional[Set[str]] = None,
    ) -> "FilteredPublisher":
        new = FilteredPublisher.__new__(cls)

        # we create a deepcopy since the aware publisher is not included in the PublisherGroup
        new.__dict__.update(publisher.__dict__)
        new.__source_types_filter__ = source_types or set()
        new.__language_filter__ = languages or set()
        return new

    @property
    def source_mapping(self) -> Dict[Type[URLSource], List[URLSource]]:
        filtered_mapping: Dict[Type[URLSource], List[URLSource]] = {}

        # iterate over internal mapping to preserve order
        for source_type, sources in self._source_mapping.items():
            if self.__source_types_filter__ and source_type not in self.__source_types_filter__:
                continue

            filtered_sources = [
                source
                for source in sources
                if not self.__language_filter__ or source.languages & self.__language_filter__
            ]
            if filtered_sources:
                filtered_mapping[source_type] = filtered_sources

        return filtered_mapping

    @property
    def language_filter(self) -> Set[str]:
        return self.__language_filter__


class PublisherGroup(type):
    def __new__(cls, name, bases, attributes):
        new = super().__new__(cls, name, bases, attributes)

        # set __name__ and __group__
        for attribute, value in attributes.items():
            if isinstance(value, Publisher):
                value.__name__ = attribute
                value.__group__ = new
                if default_language := attributes.get("default_language"):
                    for source in more_itertools.flatten(value.source_mapping.values()):
                        if not source.languages:
                            source.languages = {default_language}

        return new

    @property
    def mapping(cls) -> Dict[str, Union[Publisher, "PublisherGroup"]]:
        return {name: value for name, value in cls.__dict__.items() if isinstance(value, (Publisher, PublisherGroup))}

    def get_subgroup_mapping(cls) -> Dict[str, "PublisherGroup"]:
        return {name: value for name, value in cls.__dict__.items() if isinstance(value, PublisherGroup)}

    def __iter__(cls) -> Iterator[Publisher]:
        """This will iterate over all publishers included in the group and its subgroups.

        Returns:
            Iterator[Publisher]: Iterator over publishers included in the group and its subgroups.

        """
        for attribute in cls.__dict__.values():
            if isinstance(attribute, Publisher):
                yield attribute
            elif isinstance(attribute, PublisherGroup):
                yield from attribute

    def __getitem__(cls, name: str) -> Publisher:
        """Get a publisher from the collection by name represented as string.

        Args:
            name: A string referencing the publisher in the corresponding enum.

        Returns:
            Publisher: The corresponding publisher.

        """
        return {publisher.__name__: publisher for publisher in cls}[name]

    def __len__(cls) -> int:
        """The number of publishers included in the group.

        Returns:
            int: The number of publishers.
        """
        return len(list(cls.__iter__()))

    def __str__(cls) -> str:
        representation = f"<{cls.__name__}: {len(cls)}>"
        for name, element in cls.mapping.items():
            representation += "\n" + indent(str(element), prefix="\t")
        return representation

    def search(
        cls,
        attributes: Optional[List[str]] = None,
        source_types: Optional[List[Type[URLSource]]] = None,
        languages: Optional[List[str]] = None,
        include_deprecated_attributes: bool = False,
    ) -> Union[List[FilteredPublisher], List[Publisher]]:
        if not (attributes or source_types or languages):
            raise ValueError("You have to define at least one search condition")
        if not attributes:
            attributes = []
        if not languages:
            languages = []
        if not source_types:
            source_types = []

        matched: Union[List[FilteredPublisher], List[Publisher]] = []
        unique_attributes = set(attributes)
        spec: Publisher
        for publisher in cls:
            if unique_attributes.issubset(
                set(publisher.parser().attributes().names)
                - (
                    set(
                        publisher.parser().attributes().deprecated.names if not include_deprecated_attributes else set()
                    )
                )
            ) and (publisher.supports(source_types=source_types, languages=languages)):
                matched.append(
                    FilteredPublisher.from_publisher(
                        publisher,
                        source_types=set(source_types) & publisher.source_types,
                        languages=set(languages) & publisher.languages,
                    )
                )
        if not matched:
            warn("No publisher found matching the search criteria. Returning no publishers.")
        return matched



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.scraping.url import NewsMap, RSSFeed, Sitemap

from .derstandard import DerStandardParser
from .die_presse import DiePresseParser
from .kleine_zeitung import KleineZeitungParser
from .orf import OrfParser
from .salzburger_nachrichten import SalzburgerNachrichtenParser

# noinspection PyPep8Naming


class AT(metaclass=PublisherGroup):
    default_language = "de"

    DerStandard = Publisher(
        name="Der Standard",
        domain="https://derstandard.at",
        parser=DerStandardParser,
        sources=[
            RSSFeed("https://www.derstandard.at/rss"),
            NewsMap("https://www.derstandard.at/sitemaps/news.xml"),
            Sitemap("https://www.derstandard.at/sitemaps/sitemap.xml"),
        ],
        request_header={"user-agent": "Googlebot"},
    )

    DiePresse = Publisher(
        name="Die Presse",
        domain="https://diepresse.com",
        parser=DiePresseParser,
        sources=[
            NewsMap("https://www.diepresse.com/news-sitemap"),
        ],
    )

    KleineZeitung = Publisher(
        name="Kleine Zeitung",
        domain="https://www.kleinezeitung.at",
        parser=KleineZeitungParser,
        sources=[
            NewsMap("https://www.kleinezeitung.at/news-sitemap"),
            Sitemap("https://www.kleinezeitung.at/sitemaps/sitemap_main.xml", reverse=True),
        ],
    )

    ORF = Publisher(
        name="Österreichischer Rundfunk",
        domain="https://www.orf.at",
        parser=OrfParser,
        sources=[RSSFeed("https://rss.orf.at/news.xml")],
    )

    SalzburgerNachrichten = Publisher(
        name="Salzburger Nachrichten",
        domain="https://www.sn.at",
        parser=SalzburgerNachrichtenParser,  # Placeholder for future parser
        sources=[
            NewsMap("https://www.sn.at/sitemap/2/0001.xml"),
            Sitemap("https://www.sn.at/sitemapindex.xml", reverse=True),
        ],
    )



import functools
import inspect
import itertools
import re
from abc import ABC
from copy import copy
from dataclasses import dataclass, field
from datetime import date, datetime
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

import lxml.html

from fundus.logging import create_logger
from fundus.parser.data import LinkedDataMapping
from fundus.parser.utility import get_ld_content, get_meta_content

RegisteredFunctionT_co = TypeVar("RegisteredFunctionT_co", covariant=True, bound="RegisteredFunction")

logger = create_logger(__name__)


class RegisteredFunction(ABC):
    __wrapped__: Callable[[object], Any]
    __name__: str
    __func__: Callable[[object], Any]
    __self__: Optional["BaseParser"]

    # TODO: ensure uint for priority instead of int
    def __init__(self, func: Callable[[object], Any], priority: Optional[int]):
        self.__self__ = None
        self.__func__ = func
        self.__finite__: bool = False

        self.priority = priority

    def __get__(self, instance, owner):
        if instance and not self.__self__:
            method = copy(self)
            method.__self__ = instance
            method.__finite__ = True
            return method
        return self

    def __call__(self):
        if self.__self__ and hasattr(self.__self__, "precomputed"):
            return self.__func__(self.__self__)
        else:
            raise ValueError("You are not allowed to call attributes or functions outside the parse() method")

    def __lt__(self, other):
        if self.priority is None:
            return False
        elif other.priority is None:
            return True
        else:
            return self.priority < other.priority

    def __repr__(self):
        if instance := self.__self__:
            return f"bound {type(self).__name__} {self.__name__!r} of {instance}"
        else:
            return f"registered {type(self).__name__} {self.__name__!r}"


class Attribute(RegisteredFunction):
    def __init__(
        self,
        func: Callable[[object], Any],
        priority: Optional[int],
        validate: bool,
        deprecated: Optional[date],
        default_factory: Optional[Callable[[], Any]],
    ):
        self.validate = validate
        self.deprecated = deprecated
        self.default_factory = default_factory
        super(Attribute, self).__init__(func=func, priority=priority)

    @functools.cached_property
    def __default__(self):
        if self.default_factory is not None:
            return self.default_factory()

        annotation = self.__annotations__["return"]
        origin = get_origin(annotation)
        args = get_args(annotation)

        if not (origin or args):
            default = annotation()
        elif origin == Union:
            if type(None) in args:
                default = None
            else:
                raise NotImplementedError(f"Cannot determine default for {origin!r} with args {args!r}")
        elif isinstance(origin, type):
            default = origin()
        else:
            raise NotImplementedError(f"Unsupported origin {origin}")
        return default


class Function(RegisteredFunction):
    def __init__(self, func: Callable[[object], Any], priority: Optional[int]):
        super(Function, self).__init__(func=func, priority=priority)


def _register(cls, factory: Type[RegisteredFunction], **kwargs):
    def wrapper(func):
        return functools.update_wrapper(factory(func, **kwargs), func)

    # _register was called with parenthesis
    if cls is None:
        return wrapper

    # register was called without parenthesis
    return wrapper(cls)


def attribute(
    cls=None,
    /,
    *,
    priority: Optional[int] = None,
    validate: bool = True,
    deprecated: Optional[date] = None,
    default_factory: Optional[Callable[[], Any]] = None,
):
    return _register(
        cls,
        factory=Attribute,
        priority=priority,
        validate=validate,
        deprecated=deprecated,
        default_factory=default_factory,
    )


def function(cls=None, /, *, priority: Optional[int] = None):
    return _register(cls, factory=Function, priority=priority)


class RegisteredFunctionCollection(Collection[RegisteredFunctionT_co]):
    def __init__(self, *functions: RegisteredFunctionT_co):
        self.functions = tuple(functions)

    @property
    def names(self) -> List[str]:
        return [func.__name__ for func in self.functions]

    def __len__(self) -> int:
        return len(self.functions)

    def __iter__(self) -> Iterator[RegisteredFunctionT_co]:
        return iter(self.functions)

    def __contains__(self, item) -> bool:
        return self.functions.__contains__(item)

    def __eq__(self, other: object) -> bool:
        return self.functions == other.functions if isinstance(other, RegisteredFunctionCollection) else False

    def __str__(self) -> str:
        return ", ".join(self.names)


class AttributeCollection(RegisteredFunctionCollection[Attribute]):
    @property
    def validated(self) -> "AttributeCollection":
        return AttributeCollection(*[attr for attr in self.functions if attr.validate])

    @property
    def unvalidated(self) -> "AttributeCollection":
        return AttributeCollection(*[attr for attr in self.functions if not attr.validate])

    @property
    def deprecated(self) -> "AttributeCollection":
        return AttributeCollection(*[attr for attr in self.functions if attr.deprecated is not None])


class FunctionCollection(RegisteredFunctionCollection[Function]):
    pass


@dataclass
class Precomputed:
    html: str
    doc: lxml.html.HtmlElement
    meta: Dict[str, str]
    ld: LinkedDataMapping
    cache: Dict[str, Any] = field(default_factory=dict)


class BaseParser(ABC):
    VALID_UNTIL: date = date.today()
    precomputed: Precomputed

    def __init__(self, timestamp: Optional[date] = None):
        def predicate(x: object) -> bool:
            if not isinstance(x, RegisteredFunction):
                return False
            if isinstance(x, Attribute):
                if timestamp is None or x.deprecated is None or x.deprecated > timestamp:
                    return True
                else:
                    return False
            return True

        predicated_members: List[Tuple[str, RegisteredFunction]] = inspect.getmembers(self, predicate=predicate)
        bound_registered_functions: List[RegisteredFunction] = [func for _, func in predicated_members]
        self._sorted_registered_functions = sorted(bound_registered_functions, key=lambda f: (f, f.__name__))

    @classmethod
    def _search_members(cls, obj_type: type) -> List[Tuple[str, Any]]:
        members = inspect.getmembers(cls, predicate=lambda x: isinstance(x, obj_type)) if obj_type else []
        return members

    @classmethod
    def attributes(cls, include_all: bool = False) -> AttributeCollection:
        # We exclude both __ld and __meta here to hide them from the autogenerated tables. We do so, because
        # we neither want them to appear in the attribute_guidelines nor the supported_publishers tables.
        attrs: List[Attribute] = [
            func for _, func in cls._search_members(Attribute) if include_all or func.__name__ not in ["__ld", "__meta"]
        ]
        return AttributeCollection(*attrs)

    @classmethod
    def functions(cls) -> FunctionCollection:
        funcs: List[Function] = [func for _, func in cls._search_members(Function)]
        return FunctionCollection(*funcs)

    @property
    def cache(self) -> Optional[Dict[str, Any]]:
        return self.precomputed.cache if self.precomputed else None

    @property
    def registered(self) -> List[RegisteredFunction]:
        return self._sorted_registered_functions

    def _base_setup(self, html: str) -> None:
        doc = lxml.html.document_fromstring(html)
        self.precomputed = Precomputed(html, doc, get_meta_content(doc), get_ld_content(doc))

    def parse(self, html: str, error_handling: Literal["suppress", "catch", "raise"] = "raise") -> Dict[str, Any]:
        # wipe existing precomputed
        self._base_setup(html)

        parsed_data = {}

        for func in self._sorted_registered_functions:
            attribute_name = re.sub(r"^_{1,2}([^_]*_?)$", r"\g<1>", func.__name__)

            if isinstance(func, Function):
                func()

            elif isinstance(func, Attribute):
                try:
                    parsed_data[attribute_name] = func()
                except Exception as err:
                    if error_handling == "suppress":
                        parsed_data[attribute_name] = func.__default__
                        logger.info(
                            f"Couldn't parse attribute {attribute_name!r} for "
                            f"{self.precomputed.meta.get('og:url')!r}: {err}"
                        )
                    elif error_handling == "catch":
                        parsed_data[attribute_name] = err
                    elif error_handling == "raise":
                        raise err
                    else:
                        raise ValueError(f"Invalid value {error_handling!r} for parameter <error_handling>")

            else:
                raise TypeError(f"Invalid type for {func}. Only subclasses of 'RegisteredFunction' are allowed")

        return parsed_data

    def share(self, **kwargs):
        for key, value in kwargs.items():
            self.precomputed.cache[key] = value

    # base attribute section
    @attribute
    def __meta(self) -> Dict[str, Any]:
        return self.precomputed.meta

    @attribute
    def __ld(self) -> Optional[LinkedDataMapping]:
        return self.precomputed.ld

    @attribute
    def free_access(self) -> bool:
        if (isAccessibleForFree := self.precomputed.ld.bf_search("isAccessibleForFree")) is None:
            return True
        elif not isAccessibleForFree or isAccessibleForFree == "false" or isAccessibleForFree == "False":
            return False
        else:
            return True


class _ParserCache:
    def __init__(self, factory: Type[BaseParser]):
        self.factory: Type[BaseParser] = factory
        self._instances: Dict[Optional[date], BaseParser] = {}
        deprecated_dates = sorted(
            [attr.deprecated for attr in self.factory.attributes() if attr.deprecated is not None]
        )
        self._deprecation_timeframes: List[date] = deprecated_dates

    def __call__(self, timestamp: Optional[date] = None) -> BaseParser:
        effective_timestamp: Optional[date] = None
        if timestamp:
            for timeframe_date in reversed(self._deprecation_timeframes):
                if timeframe_date <= timestamp:
                    effective_timestamp = timeframe_date
                    break
        if effective_timestamp not in self._instances:
            self._instances[effective_timestamp] = self.factory(timestamp=effective_timestamp)

        return self._instances[effective_timestamp]


class ParserProxy(ABC):
    def __init__(self):
        predicate: Callable[[object], bool] = lambda x: inspect.isclass(x) and issubclass(x, BaseParser)
        included_parsers: List[Type[BaseParser]] = [
            parser for name, parser in inspect.getmembers(type(self), predicate=predicate)
        ]

        if not included_parsers:
            raise ValueError(
                f"<class {type(self).__name__}> consists of no parser-versions. "
                f"To include versions add subclasses of <class {BaseParser.__name__}> to the class definition."
            )

        mapping: Dict[date, _ParserCache] = {}
        for versioned_parser in sorted(included_parsers, key=lambda parser: parser.VALID_UNTIL):
            validation_date: date
            if prev := mapping.get(validation_date := versioned_parser.VALID_UNTIL):  # type: ignore
                raise ValueError(
                    f"Found versions {prev.factory.__name__!r} and {versioned_parser.__name__!r} of "
                    f"{str(self)!r} with same validation date.\nMake sure you use class attribute VALID_UNTIL "
                    f"of <class {BaseParser.__name__}> to set validation dates for legacy versions."
                )
            mapping[validation_date] = _ParserCache(versioned_parser)
        self._parser_mapping = mapping

    def __call__(self, crawl_date: Optional[Union[datetime, date]] = None) -> BaseParser:
        if crawl_date is None:
            return self._get_latest_cache()()

        parsed_date = crawl_date.date() if isinstance(crawl_date, datetime) else crawl_date
        parser_cache: _ParserCache
        try:
            _, parser_cache = next(itertools.dropwhile(lambda x: x[0] < parsed_date, self._parser_mapping.items()))
        except StopIteration:
            raise ValueError(
                f"Couldn't find a fitting parser valid at date {parsed_date}. "
                f"Last valid date is {self._get_latest_cache()().VALID_UNTIL}"
            )
        return parser_cache(parsed_date)

    def __iter__(self) -> Iterator[Type[BaseParser]]:
        """Iterates over all included parser versions with the latest being first.

        Returns:
            Iterator over included parser versions
        """
        return (cache.factory for cache in reversed(self._parser_mapping.values()))

    def __len__(self) -> int:
        return len(self._parser_mapping)

    def __bool__(self) -> bool:
        return bool(self._parser_mapping)

    def __str__(self) -> str:
        return f"<{ParserProxy.__name__} {type(self).__name__}>"

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__} including versions "
            f"{', '.join([cache.factory.__name__ for cache in self._parser_mapping.values()])!r}"
            if self._parser_mapping
            else f"Empty {type(self).__name__}"
        )

    @property
    def attribute_mapping(self) -> Dict[Type[BaseParser], AttributeCollection]:
        return {versioned_parser: versioned_parser.attributes() for versioned_parser in self}

    @property
    def function_mapping(self) -> Dict[Type[BaseParser], FunctionCollection]:
        return {versioned_parser: versioned_parser.functions() for versioned_parser in self}

    def _get_latest_cache(self) -> _ParserCache:
        return list(self._parser_mapping.values())[-1]

    @property
    def latest_version(self) -> Type[BaseParser]:
        return self._get_latest_cache().factory



import datetime
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class BusinessInsiderParser(ParserProxy):
    class V1(BaseParser):
        VALID_UNTIL = datetime.date(2025, 3, 24)

        _summary_selector = CSSSelector("article ul[class^='summary-list'] > li")
        _subheadline_selector = CSSSelector("article h2, div.slideshow-slide-container h2")
        _paragraph_selector = XPath(
            """
            //article 
            //div[contains(@class, 'content-lock-content')] 
            /p[not(contains(@class, 'disclaimer'))] | 
            //article 
            //div[contains(@class, 'content-lock-content')]
            /div[contains(@class, 'premium-content')] 
            /p[not(contains(@class, 'disclaimer'))] | 
            //div[@class='slide-layout clearfix']
            /p[not(contains(@class, 'disclaimer'))]
            """
        )

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.bf_search("author"))

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("title")

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(
                self.precomputed.meta.get("keywords")
                or self.precomputed.ld.bf_search("keywords")
                or self.precomputed.meta.get("news_keywords")
            )

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=XPath("//article"),
                image_selector=XPath("//figure//img[not(@data-content-type)]"),
                caption_selector=XPath("./ancestor::figure//figcaption/span[@class='image-caption-text']"),
                author_selector=XPath("./ancestor::figure//figcaption/span[@class='image-source-text']"),
            )

    class V2(BaseParser):
        VALID_UNTIL = datetime.date.today()

        _paragraph_selector = XPath("//section[contains(@class, 'post-body-content')]/p")
        _summary_selector = XPath("//div[@class='post-summary-bullets']//li")
        _subheadline_selector = XPath("//section[contains(@class, 'post-body-content')]/h2")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.bf_search("author"))

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("title")

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(
                self.precomputed.meta.get("keywords")
                or self.precomputed.ld.bf_search("keywords")
                or self.precomputed.meta.get("news_keywords")
            )



from datetime import date, datetime

from dateutil.rrule import YEARLY, rrule

from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.scraping.filter import inverse, regex_filter
from fundus.scraping.url import NewsMap, RSSFeed, Sitemap

from ..shared import EuronewsParser
from .daily_mail import DailyMailParser
from .daily_star import DailyStarParser
from .evening_standard import EveningStandardParser
from .express import ExpressParser
from .i_news import INewsParser
from .metro import MetroParser
from .the_bbc import TheBBCParser
from .the_guardian import TheGuardianParser
from .the_independent import TheIndependentParser
from .the_mirror import TheMirrorParser
from .the_sun import TheSunParser
from .the_telegraph import TheTelegraphParser


class UK(metaclass=PublisherGroup):
    default_language = "en"

    TheGuardian = Publisher(
        name="The Guardian",
        domain="https://www.theguardian.com/",
        parser=TheGuardianParser,
        sources=[NewsMap("https://www.theguardian.com/sitemaps/news.xml")],
    )

    TheIndependent = Publisher(
        name="The Independent",
        domain="https://www.independent.co.uk/",
        parser=TheIndependentParser,
        sources=[
            Sitemap(
                "https://www.independent.co.uk/sitemap.xml",
                sitemap_filter=inverse(regex_filter(f"sitemap-articles")),
            ),
            NewsMap("https://www.independent.co.uk/sitemaps/googlenews"),
        ],
    )

    TheMirror = Publisher(
        name="The Mirror",
        domain="https://www.mirror.co.uk/",
        parser=TheMirrorParser,
        sources=[
            Sitemap("https://www.mirror.co.uk/sitemaps/sitemap_index.xml", reverse=True),
            NewsMap("https://www.mirror.co.uk/map_news.xml"),
        ],
    )

    TheTelegraph = Publisher(
        name="The Telegraph",
        domain="https://www.telegraph.co.uk/",
        parser=TheTelegraphParser,
        sources=[
            Sitemap("https://www.telegraph.co.uk/sitemap.xml"),
            NewsMap("https://www.telegraph.co.uk/custom/daily-news/sitemap.xml"),
        ],
        deprecated=True,
    )

    iNews = Publisher(
        name="i",
        domain="https://inews.co.uk/",
        parser=INewsParser,
        sources=[
            Sitemap("https://inews.co.uk/sitemap.xml"),
            NewsMap(
                f"https://inews.co.uk/sitemap.xml"
                f"?yyyy={date.today().year}&mm={str(date.today().month).zfill(2)}&dd={str(date.today().day).zfill(2)}",
            ),
        ],
    )

    EuronewsEN = Publisher(
        name="Euronews (EN)",
        domain="https://www.euronews.com/",
        parser=EuronewsParser,
        sources=[
            Sitemap("https://www.euronews.com/sitemaps/en/articles.xml"),
            NewsMap("https://www.euronews.com/sitemaps/en/latest-news.xml"),
        ],
    )

    DailyStar = Publisher(
        name="Daily Star",
        domain="https://www.dailystar.co.uk/",
        parser=DailyStarParser,
        sources=[
            Sitemap("https://www.dailystar.co.uk/sitemaps/sitemap_index.xml", reverse=True),
            NewsMap("https://www.dailystar.co.uk/map_news.xml"),
        ],
    )

    TheSun = Publisher(
        name="The Sun",
        domain="https://www.thesun.co.uk/",
        parser=TheSunParser,
        sources=[
            Sitemap("https://www.thesun.co.uk/sitemap.xml"),
            NewsMap("https://www.thesun.co.uk/news-sitemap.xml"),
        ],
        url_filter=regex_filter("sun-bingo|web-stories"),
    )

    DailyMail = Publisher(
        name="Daily Mail",
        domain="https://www.dailymail.co.uk/",
        parser=DailyMailParser,
        sources=[
            NewsMap("https://www.dailymail.co.uk/google-news-sitemap.xml"),
        ]
        + [
            Sitemap(f"https://www.dailymail.co.uk/sitemap-articles-year~{year.year}.xml")
            for year in rrule(YEARLY, dtstart=datetime(2021, 1, 1), until=datetime.today())
        ],
    )

    EveningStandard = Publisher(
        name="Evening Standard",
        domain="https://www.standard.co.uk/",
        parser=EveningStandardParser,
        sources=[
            Sitemap(
                "https://www.standard.co.uk/sitemap.xml",
                sitemap_filter=inverse(regex_filter("sitemap-articles|sitemap-recent")),
            ),
            RSSFeed("https://www.standard.co.uk/rss"),
        ],
    )

    Metro = Publisher(
        name="Metro",
        domain="https://metro.co.uk/",
        parser=MetroParser,
        sources=[
            NewsMap("https://metro.co.uk/news-sitemap.xml"),
            Sitemap("https://metro.co.uk/sitemap.xml"),
        ],
    )

    Express = Publisher(
        name="Daily Express",
        domain="https://www.express.co.uk/",
        parser=ExpressParser,
        sources=[
            NewsMap("https://www.express.co.uk/googlenews.xml"),
            Sitemap("https://www.express.co.uk/sitemap.xml"),
        ],
    )

    BBC = Publisher(
        name="The BBC",
        domain="https://www.bbc.co.uk/",
        parser=TheBBCParser,
        sources=[
            NewsMap("https://www.bbc.co.uk/sitemaps/https-index-uk-news.xml"),
            Sitemap("https://www.bbc.co.uk/sitemaps/https-index-com-archive.xml", reverse=True),
        ],
        url_filter=regex_filter("video|live"),
    )



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.scraping.filter import inverse, regex_filter
from fundus.scraping.url import Sitemap

from .daily_news_tz import DailyNewsTZParser


class TZ(metaclass=PublisherGroup):
    default_language = "sw"

    DailyNewsTZ = Publisher(
        name="Daily News (Tanzania)",
        domain="https://www.dailynews.co.tz/",
        parser=DailyNewsTZParser,
        sources=[
            Sitemap(
                "https://dailynews.co.tz/wp-sitemap.xml",
                sitemap_filter=inverse(regex_filter("post-sitemap")),
                languages={"en"},
            ),
        ],
    )
    HabariLeo = Publisher(
        name="Habari Leo",
        domain="https://www.habarileo.co.tz/",
        parser=DailyNewsTZParser,
        sources=[
            Sitemap(
                "https://habarileo.co.tz/sitemap.xml",
                sitemap_filter=inverse(regex_filter("post-sitemap")),
            ),
        ],
    )



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.publishers.pl.rzeczpospolita import RzeczpospolitaParser
from fundus.scraping.url import NewsMap, Sitemap


class PL(metaclass=PublisherGroup):
    default_language = "pl"

    Rzeczpospolita = Publisher(
        name="Rzeczpospolita",
        domain="https://www.rp.pl/",
        parser=RzeczpospolitaParser,
        sources=[
            Sitemap(
                "https://www.rp.pl/sitemaps/sitemap.xml",
            ),
            NewsMap("https://www.rp.pl/sitemaps/news-sitemap.xml"),
        ],
    )



import datetime
import re
from typing import List, Optional

from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_date_parsing,
    image_extraction,
)


class TageszeitungParser(ParserProxy):
    class V1(BaseParser):
        _summary_selector = XPath("//div[@id='article_content']//p[not(@class='wp-caption-text' or text()) and strong]")
        _paragraph_selector = XPath("//div[@id='article_content']//p[not(@class='wp-caption-text') and text()]")

        _date_selector = XPath("//span[@class='meta_date']//strong/text()")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def title(self) -> Optional[str]:
            return re.sub(
                r"(?i)\s*-\s*Die Neue Südtiroler Tageszeitung$", "", self.precomputed.meta.get("og:title") or ""
            )

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            if not (publishing_date := self._date_selector(self.precomputed.doc)):
                return None
            else:
                return generic_date_parsing(publishing_date[0])

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=XPath("//div[@id='article_content']"),
                image_selector=XPath("//article//img"),
                caption_selector=XPath("./ancestor::div[@class='wp-caption alignnone']//p[@class='wp-caption-text']"),
                author_selector=re.compile(r"(^|\()(Fotos?:|©)(?P<credits>[^)]+)\)?"),
            )



import _thread as thread
import contextlib
import threading
import time
from typing import Callable, Iterator, Optional

from typing_extensions import ParamSpec

P = ParamSpec("P")


class Stopwatch:
    def __init__(self):
        self._start = time.time()

    @property
    def time(self) -> float:
        return max(0.0, time.time() - self._start)

    def reset(self):
        self._start = time.time()


class ResettableTimer(threading.Thread):
    def __init__(
        self,
        seconds: float,
        func: Callable[P, None],
        interval: float = 0.1,
        args: P.args = tuple(),
        kwargs: P.kwargs = None,
    ) -> None:
        """Resettable timer executing <func> after <time> seconds, checking every <interval>.

        Args:
            seconds: Time to pass in seconds.
            func: Callable to execute when <time> has passed.
            interval: Check every <interval> seconds if condition is met (reduce workload on CPU).
            *args: Arguments to <func>.
            **kwargs: Keyword arguments to <func>.
        """
        super().__init__(target=func, args=args, kwargs=kwargs)
        self.seconds = seconds
        self.interval = interval
        self.watch = Stopwatch()
        self._canceled = threading.Event()

    def run(self) -> None:
        self.watch.reset()
        while True and self.watch.time < self.seconds:
            time.sleep(self.interval)
            if self._canceled.is_set():
                return
        # noinspection PyUnresolvedReferences
        self._target(*self._args, **self._kwargs)  # type: ignore[attr-defined]

    def reset(self) -> None:
        self.watch.reset()

    def cancel(self) -> None:
        self._canceled.set()


def _interrupt_handler() -> None:
    thread.interrupt_main()


# noinspection PyPep8Naming
@contextlib.contextmanager
def Timeout(
    seconds: float, silent: bool = False, callback: Optional[Callable[[], None]] = None, disable: bool = False
) -> Iterator[ResettableTimer]:
    """Context manager applying a resettable timeout.

    Contextmanager implementation of timeout which does not relly on a function.
    If enter the context manager will time out after <time> seconds.
    See docstring of 'timeout' for more information

    Args:
        seconds: The time after which to timout in seconds. If set to <= 0, set timer never start.
        silent: If True, the KeyboardInterrupt will be silently ignored and None returned instead.
            Defaults to False.
        callback: If given, will be called instead of raising KeyboardInterrupt. Defaults to None.
        disable: If True, the timer will never start effectively disable the timeout.

    Returns:
        ResettableTimer: A timer to reset or cancel the timeout.
    """
    timer = ResettableTimer(seconds, callback or _interrupt_handler)
    try:
        if not disable:
            timer.start()
        yield timer
    except KeyboardInterrupt as err:
        if not silent:
            raise TimeoutError from err
    finally:
        timer.cancel()
        del timer



import datetime
import re
from typing import List, Optional

from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    image_extraction,
)


class TaipeiTimesParser(ParserProxy):
    class V1(BaseParser):
        _paragraph_selector = XPath(
            r"//div[@class='archives']/p[not(re:test(text(), '(?i)^（by.*）\s*$'))]",
            namespaces={"re": "http://exslt.org/regular-expressions"},
        )
        _summary_selector = XPath("//div[@class='archives']/h2")
        _author_selector = XPath("//div[@class='archives']//div[@class='name']/text()")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                summary_selector=self._summary_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            author_selection = self._author_selector(self.precomputed.doc)
            if not author_selection:
                return []
            else:
                selection = re.sub(
                    r"(?is)(^by|/.*|staff reporter|(,?\s*with\s*)?staff writer.*)", "", author_selection[0]
                )
            return generic_author_parsing(selection, split_on=[r"\s+and\s+"])

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.bf_search("headline")

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=XPath("//div[@class='archives']"),
                image_selector=XPath("//div[@class='imgboxa']//img"),
                caption_selector=XPath("./ancestor::div[@class='imgboxa']//h1"),
                author_selector=XPath("./ancestor::div[@class='imgboxa']//p"),
            )



import re
from datetime import datetime
from typing import List, Optional

from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class CorriereDellaSeraParser(ParserProxy):
    class V1(BaseParser):
        # Selectors for article body parts
        _summary_selector = XPath("//p[contains(@class, 'summary')]")
        _paragraph_selector = XPath("//p[@class='chapter-paragraph' and text()]")
        _subheadline_selector = XPath("//h2[contains(@class, 'native-summary-content')]")

        @attribute
        def title(self) -> Optional[str]:
            # Get the headline from og:title meta tag
            return self.precomputed.ld.bf_search("headline")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            # Extract article body using utility function
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                paragraph_selector=self._paragraph_selector,
                subheadline_selector=self._subheadline_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            # Extract authors from schema.org NewsArticle data
            authors = self.precomputed.ld.xpath_search("//NewsArticle/author")
            if authors:
                return generic_author_parsing(authors)
            return []

        @attribute
        def publishing_date(self) -> Optional[datetime]:
            # Use scalar parameter for direct value
            date_str = self.precomputed.ld.xpath_search("//NewsArticle/datePublished", scalar=True)
            return generic_date_parsing(date_str)

        @attribute
        def topics(self) -> List[str]:
            breadcrumb_items = self.precomputed.ld.xpath_search("//BreadcrumbList/itemListElement/*/name")
            if breadcrumb_items:
                return generic_topic_parsing(breadcrumb_items[1:])
            section = self.precomputed.ld.xpath_search("//NewsArticle/articleSection", scalar=True)
            if section:
                return generic_topic_parsing([section])
            return []

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                author_selector=re.compile(r"\(foto (?P<credits>.*)\)\s*$"),
            )



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.scraping.filter import inverse, regex_filter
from fundus.scraping.url import NewsMap, Sitemap

from .hankook_ilbo import HankookIlboParser
from .mbn import MBNParser


class KR(metaclass=PublisherGroup):
    default_language = "ko"

    MBN = Publisher(
        name="Maeil Business Newspaper",
        domain="https://www.mk.co.kr/",
        parser=MBNParser,
        sources=[
            # RSSFeed("https://www.mk.co.kr/rss/40300001/"),
            Sitemap(
                "https://www.mk.co.kr/sitemap/latest-articles/",
                sitemap_filter=inverse(regex_filter(r"/news/columnists/")),
            ),
            Sitemap(
                "https://www.mk.co.kr/sitemap/daily-articles/",
                sitemap_filter=inverse(regex_filter(r"/news/columnists/")),
            ),
        ],
    )

    HankookIlbo = Publisher(
        name="Hankook Ilbo",
        domain="https://www.hankookilbo.com/",
        parser=HankookIlboParser,
        sources=[
            NewsMap("https://www.hankookilbo.com/sitemap/latest-articles"),
            Sitemap("https://www.hankookilbo.com/sitemap/daily-articles/2020"),
        ],
    )



import datetime
from typing import List, Optional

from lxml.cssselect import CSSSelector

from fundus.parser import ArticleBody, BaseParser, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
)


class OccupyDemocratsParser(ParserProxy):
    class V1(BaseParser):
        _paragraph_selector: CSSSelector = CSSSelector(
            "div[itemprop='articleBody']>p, div[itemprop='articleBody']>blockquote"
        )

        @attribute
        def body(self) -> Optional[ArticleBody]:
            body: ArticleBody = extract_article_body_with_selector(
                self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
            )
            return body

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.bf_search("Person"))

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.meta.get("article:published_time"))

        @attribute
        def title(self) -> Optional[str]:
            title: Optional[str] = self.precomputed.meta.get("og:title")
            return title

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.ld.xpath_search("Article/keywords", scalar=True))

        @attribute(validate=False)
        def description(self) -> Optional[str]:
            return self.precomputed.meta.get("description")
